////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// SSLGuardian.cpp                                                            //
// (c) 2008 Matthias Withopf / Heise Zeitschriften Verlag                     //
//   http://www.heise.de/ct/projekte/sslwaechter/                             //
//   http://www.withopf.com/tools/sslguardian/                                //
//                                                                            //
//  This file is part of SSLGuardian.                                         //
//                                                                            //
//  SSLGuardian is free software: you can redistribute it and/or modify       //
//  it under the terms of the GNU General Public License as published by      //
//  the Free Software Foundation, either version 3 of the License, or         //
//  (at your option) any later version.                                       //
//                                                                            //
//  SSLGuardian is distributed in the hope that it will be useful,            //
//  but WITHOUT ANY WARRANTY; without even the implied warranty of            //
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             //
//  GNU General Public License for more details.                              //
//                                                                            //
//  You should have received a copy of the GNU General Public License         //
//  along with SSLGuardian.  If not, see <http://www.gnu.org/licenses/>.      //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

#include "SSLGuardian.h"

const PChar SignonStr = (PChar)"SSLGuardian V1.1 (c) 2008 Matthias Withopf / Heise Zeitschriften Verlag";

#define STRICT
#define WIN32_LEAN_AND_MEAN
#pragma warning(disable:4201)
#include <windows.h>
#include <wincrypt.h>
#include <winhttp.h>
#include <shellapi.h>
#include <strsafe.h>
#pragma warning(default:4201)
#include "utils.h"
#include "SSLGuardianMsg.h"

HINSTANCE        DllInst            = NULL;
Boolean          IsInteractive      = False;
CRITICAL_SECTION SHA1CheckCacheCS;
const uint       SHA1CheckCacheSize = 1000;
typedef struct
  {
    Char    SHA1[40 + 1];
    Boolean CheckResult;
  } TSHA1CheckCacheEntry,* PSHA1CheckCacheEntry;
PSHA1CheckCacheEntry SHA1CheckCache      = NULL;
uint                 SHA1CheckCacheCount = 0;

// Init ////////////////////////////////////////////////////////////////////////

Boolean Init(void)
{
  Boolean Result = True;

  HWINSTA W = GetProcessWindowStation();
  if (W)
    {
      Char  tmp[128];
      DWORD R;
      if (GetUserObjectInformation(W,UOI_NAME,(Pchar)tmp,sizeof tmp,&R))
        {
          if (!_strcmpi((Pchar)tmp,"Winsta0"))
            {
              USEROBJECTFLAGS UOF;
              if (GetUserObjectInformation(W,UOI_FLAGS,&UOF,sizeof UOF,&R))
                {
                  if (UOF.dwFlags & WSF_VISIBLE)
                    IsInteractive = True;
                }
            }
        }
    }

  InitLangResOfs();
  InitializeCriticalSection(&SHA1CheckCacheCS);
  SHA1CheckCache = new TSHA1CheckCacheEntry[SHA1CheckCacheSize];
  if (!SHA1CheckCache)
    Result = False;
  return Result;
}

// Done ////////////////////////////////////////////////////////////////////////

Boolean Done(void)
{
  Boolean Result = True;
  delete [] SHA1CheckCache;
  SHA1CheckCache = NULL;
  DeleteCriticalSection(&SHA1CheckCacheCS);
  return Result;
}

// HTTPRequest /////////////////////////////////////////////////////////////////

Boolean HTTPRequest(PChar ServerName,uint ServerPort,PChar URL,Boolean Secure,Pointer PostData,uint PostDataSize,PLongWord NativeErrorCode,LPCWSTR OptionalHeader,LPCWSTR *AcceptTypes)
{
  Boolean Result = False;
  if (NativeErrorCode)
    *NativeErrorCode = 0;
  WCHAR tmp[1024];
  HINTERNET hSession = WinHttpOpen(L"",WINHTTP_ACCESS_TYPE_DEFAULT_PROXY,WINHTTP_NO_PROXY_NAME,WINHTTP_NO_PROXY_BYPASS,0);
  if (!hSession)
    {
      if (NativeErrorCode)
        *NativeErrorCode = GetLastError();
    }
  else
    {
      uint l = (uint)strlen((Pchar)ServerName);
      if (l > (sizeof tmp / sizeof tmp[0] - 1))
        l = (sizeof tmp / sizeof tmp[0] - 1);
      uint i;
      for (i = 0;i < l;++i)
        tmp[i] = ServerName[i];
      tmp[l] = '\0';
      HINTERNET hConnect = WinHttpConnect(hSession,(LPCWSTR)tmp,(INTERNET_PORT)(ServerPort ? ServerPort : (Secure ? INTERNET_DEFAULT_HTTPS_PORT : INTERNET_DEFAULT_HTTP_PORT)),0);
      if (!hConnect)
        {
          if (NativeErrorCode)
            *NativeErrorCode = GetLastError();
        }
      else
        {
          uint l = (uint)strlen((Pchar)URL);
          if (l > (sizeof tmp / sizeof tmp[0] - 1))
            l = (sizeof tmp / sizeof tmp[0] - 1);
          for (i = 0;i < l;++i)
            tmp[i] = URL[i];
          tmp[l] = '\0';
          HINTERNET hRequest = WinHttpOpenRequest(hConnect,PostData ? L"POST" : L"GET",(LPCWSTR)tmp,NULL,WINHTTP_NO_REFERER,AcceptTypes ? AcceptTypes : WINHTTP_DEFAULT_ACCEPT_TYPES,Secure ? WINHTTP_FLAG_SECURE : 0);
          if (hRequest)
            {
              BOOL f = WinHttpSendRequest(hRequest,OptionalHeader ? OptionalHeader : WINHTTP_NO_ADDITIONAL_HEADERS,0,PostData ? PostData : WINHTTP_NO_REQUEST_DATA,PostData ? PostDataSize : 0,PostData ? PostDataSize : 0,0);
              if (!f)
                {
                  if (NativeErrorCode)
                    *NativeErrorCode = GetLastError();
                }
              else
                {
                  f = WinHttpReceiveResponse(hRequest,NULL);
                  if (!f)
                    {
                      if (NativeErrorCode)
                        *NativeErrorCode = GetLastError();
                    }
                  else
                    {
                      Result = True;
                      for (;;)
                        {
                          DWORD Size = 0;
                          f = WinHttpQueryDataAvailable(hRequest,&Size);
                          if (!f)
                            {
                              if (NativeErrorCode)
                                *NativeErrorCode = GetLastError();
                              Result = False;
                              break;
                            }
                          if (!Size)
                            break;
                          LPSTR Buf = new char[Size + 1];
                          if (!Buf)
                            {
                              Result = False;
                              break;
                            }
                          DWORD Size2 = 0;
                          f = WinHttpReadData(hRequest,(LPVOID)Buf,Size,&Size2);
                          if (!f)
                            {
                              if (NativeErrorCode)
                                *NativeErrorCode = GetLastError();
                              Result = False;
                              break;
                            }
                          if (Size2)
                            {
                              Buf[Size2] = '\0';
                            }
                          delete [] Buf;
                          if (!Size2)
                            break;
                        }
                    }
                }
              WinHttpCloseHandle(hRequest);
            }
          WinHttpCloseHandle(hConnect);
        }
      WinHttpCloseHandle(hSession);
    }
  return Result;
}

// TSHA1 ///////////////////////////////////////////////////////////////////////

class TSHA1
  {
  public:
    TSHA1(void);
    ~TSHA1(void);
    Boolean Clear(void);
    Boolean Update(PByte input,uint inputLen);
    Boolean Final(void);
    enum  { DigestStrSize = 40 };
    PChar   GetDigestStr(void);
    enum  { DigestSize = 20 };
  private:
    struct _TSHA1Context *L;
    Byte digest[DigestSize];
    Char DigestStr[DigestStrSize + 1];
  };
typedef TSHA1 * PSHA1;

////////////////////////////////////////////////////////////////////////////////

enum
  {
    shaSuccess = 0,
    shaNull,
    shaInputTooLong,
    shaStateError
  };

const uint SHA1HashSize = 20;

typedef struct _TSHA1Context
  {
    LongWord Intermediate_Hash[SHA1HashSize / 4];  
    LongWord Length_Low;                           
    LongWord Length_High;                          
    Integer  Message_Block_Index;                  
    Byte     Message_Block[64];
    Boolean  Computed;                             
    int      Corrupted;                            
  } TSHA1Context,* PSHA1Context;

#define SHA1CircularShift(bits,word) (((word) << (bits)) | ((word) >> (32 - (bits))))

int SHA1Reset(PSHA1Context context)
{
  if (!context)
    return shaNull;
  context->Length_Low             = 0;
  context->Length_High            = 0;
  context->Message_Block_Index    = 0;
  context->Intermediate_Hash[0]   = 0x67452301;
  context->Intermediate_Hash[1]   = 0xEFCDAB89;
  context->Intermediate_Hash[2]   = 0x98BADCFE;
  context->Intermediate_Hash[3]   = 0x10325476;
  context->Intermediate_Hash[4]   = 0xC3D2E1F0;
  context->Computed   = False;
  context->Corrupted  = 0;
  return shaSuccess;
}

void SHA1ProcessMessageBlock(PSHA1Context context)
{
  static const LongWord K[] =
    {
      0x5A827999,
      0x6ED9EBA1,
      0x8F1BBCDC,
      0xCA62C1D6
    };
  LongWord temp;
  LongWord W[80];
  int t;
  for (t = 0;t < 16;++t)
    {
      W[t]  = context->Message_Block[t * 4] << 24;
      W[t] |= context->Message_Block[t * 4 + 1] << 16;
      W[t] |= context->Message_Block[t * 4 + 2] << 8;
      W[t] |= context->Message_Block[t * 4 + 3];
    }
  for (t = 16;t < 80;++t)
    {
      W[t] = SHA1CircularShift(1,W[t-3] ^ W[t-8] ^ W[t-14] ^ W[t-16]);
    }
  LongWord A = context->Intermediate_Hash[0];
  LongWord B = context->Intermediate_Hash[1];
  LongWord C = context->Intermediate_Hash[2];
  LongWord D = context->Intermediate_Hash[3];
  LongWord E = context->Intermediate_Hash[4];
  for (t = 0;t < 20;++t)
    {
      temp = SHA1CircularShift(5,A) + ((B & C) | ((~B) & D)) + E + W[t] + K[0];
      E = D;
      D = C;
      C = SHA1CircularShift(30,B);
      B = A;
      A = temp;
    }
  for (t = 20;t < 40;++t)
    {
      temp = SHA1CircularShift(5,A) + (B ^ C ^ D) + E + W[t] + K[1];
      E = D;
      D = C;
      C = SHA1CircularShift(30,B);
      B = A;
      A = temp;
    }
  for (t = 40; t < 60;++t)
    {
      temp = SHA1CircularShift(5,A) + ((B & C) | (B & D) | (C & D)) + E + W[t] + K[2];
      E = D;
      D = C;
      C = SHA1CircularShift(30,B);
      B = A;
      A = temp;
    }
  for (t = 60;t < 80;++t)
    {
      temp = SHA1CircularShift(5,A) + (B ^ C ^ D) + E + W[t] + K[3];
      E = D;
      D = C;
      C = SHA1CircularShift(30,B);
      B = A;
      A = temp;
    }
  context->Intermediate_Hash[0] += A;
  context->Intermediate_Hash[1] += B;
  context->Intermediate_Hash[2] += C;
  context->Intermediate_Hash[3] += D;
  context->Intermediate_Hash[4] += E;
  context->Message_Block_Index = 0;
}

void SHA1PadMessage(PSHA1Context context)
{
  if (context->Message_Block_Index > 55)
    {
      context->Message_Block[context->Message_Block_Index++] = 0x80;
      while(context->Message_Block_Index < 64)
        context->Message_Block[context->Message_Block_Index++] = 0;
      SHA1ProcessMessageBlock(context);
      while(context->Message_Block_Index < 56)
        context->Message_Block[context->Message_Block_Index++] = 0;
    }
  else
    {
      context->Message_Block[context->Message_Block_Index++] = 0x80;
      while(context->Message_Block_Index < 56)
        context->Message_Block[context->Message_Block_Index++] = 0;
    }
  context->Message_Block[56] = (Byte)(context->Length_High >> 24);
  context->Message_Block[57] = (Byte)(context->Length_High >> 16);
  context->Message_Block[58] = (Byte)(context->Length_High >> 8);
  context->Message_Block[59] = (Byte)(context->Length_High);
  context->Message_Block[60] = (Byte)(context->Length_Low >> 24);
  context->Message_Block[61] = (Byte)(context->Length_Low >> 16);
  context->Message_Block[62] = (Byte)(context->Length_Low >> 8);
  context->Message_Block[63] = (Byte)(context->Length_Low);
  SHA1ProcessMessageBlock(context);
}

int SHA1Result(PSHA1Context context,Byte Message_Digest[SHA1HashSize])
{
  if (!context || !Message_Digest)
    return shaNull;
  if (context->Corrupted)
    return context->Corrupted;
  int i;
  if (!context->Computed)
    {
      SHA1PadMessage(context);
      for (i = 0;i < 64;++i)
        context->Message_Block[i] = 0;
      context->Length_Low  = 0;
      context->Length_High = 0;
      context->Computed    = True;
    }
  for (i = 0;i < SHA1HashSize;++i)
    Message_Digest[i] = (Byte)(context->Intermediate_Hash[i >> 2] >> 8 * (3 - (i & 0x03)));
  return shaSuccess;
}

int SHA1Input(PSHA1Context context,const Byte *message_array,unsigned int length)
{
  if (!length)
    return shaSuccess;
  if (!context || !message_array)
    return shaNull;
  if (context->Computed)
    {
      context->Corrupted = shaStateError;
      return shaStateError;
    }
  if (context->Corrupted)
    return context->Corrupted;
  while (length-- && !context->Corrupted)
    {
      context->Message_Block[context->Message_Block_Index++] = (Byte)(*message_array & 0xFF);
      context->Length_Low += 8;
      if (context->Length_Low == 0)
        {
          ++context->Length_High;
          if (context->Length_High == 0)
            context->Corrupted = 1;
        }
      if (context->Message_Block_Index == 64)
        SHA1ProcessMessageBlock(context);
      ++message_array;
    }
  return shaSuccess;
}

TSHA1::TSHA1(void)
{
  L = new TSHA1Context;
  if (L)
    {
      Clear();
    }
}

TSHA1::~TSHA1(void)
{
  if (L)
    {
      delete L;
    }
}

Boolean TSHA1::Clear(void)
{
  Boolean Result = False;
  if (L)
    {
      memset(L,0,sizeof *L);
      memset(digest,0,sizeof digest);
      DigestStr[0] = '\0';
      SHA1Reset(L);
      Result = True;
    }
  return Result;
}

Boolean TSHA1::Update(PByte input,uint inputLen)
{
  Boolean Result = False;
  if (L)
    {
      int err = SHA1Input(L,(PByte)input,inputLen);
      if (!err)
        Result = True;
    }
  return Result;
}

Boolean TSHA1::Final(void)
{
  Boolean Result = False;
  if (L)
    {
      int err = SHA1Result(L,digest);
      if (!err)
        Result = True;
    }
  return Result;
}

PChar TSHA1::GetDigestStr(void)
{
  DigestStr[0] = '\0';
  for (uint i = 0;i < DigestSize;++i)
    {
      Char tmp[10];
      StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)"%02X",digest[i]);
      StringCbCat((Pchar)DigestStr,sizeof DigestStr,(Pchar)tmp);
    }
  CharLower((Pchar)DigestStr);
  return DigestStr;
}

// TProcessCertificate /////////////////////////////////////////////////////////

class TProcessCertificate
  {
  public:
    typedef enum
      {
        cec_Ok,
        cec_Uninitialized,
        cec_UnknownError,
        cec_BadInfo,
        cec_NestingTooDeep,
        cec_IdDataToBig,
        cec_NoLengthData,
        cec_IndefiniteFormNotSupported,
        cec_LengthDataTooBig,
        cec_InsufficientLengthData,
        cec_InvalidLengthData,
        cec_InvalidContentData,
        cec_ProcessEntryError,
        cec_LAST
      } TCertErrorCode;
    TCertErrorCode CertErrorCode;
    TProcessCertificate(void);
    virtual ~TProcessCertificate();
            Boolean Process(Pointer Data,uint DataSize);
    virtual Boolean ProcessEntry(LongWord Id,Byte Class,PByte Data,uint DataSize) = 0;
  private:
    Boolean DoProcess(uint Depth,PByte p,uint Size,PByte pStart,PByte pEnd);
  };

TProcessCertificate::TProcessCertificate(void)
{
  CertErrorCode = cec_Uninitialized;
}

TProcessCertificate::~TProcessCertificate()
{
}

Boolean TProcessCertificate::DoProcess(uint Depth,PByte p,uint Size,PByte pStart,PByte pEnd)
{
  Boolean Result = False;
  if (Depth > 1024)
    CertErrorCode = cec_NestingTooDeep;
  else
    {
      PByte p1 = p + Size;
      while (p < p1)
        {
          Result = True;
          Byte b = *p++;
          Byte     IdentifierClass = (Byte)((b >> 6) & 3);
          Byte     IdentifierPC    = (Byte)((b >> 5) & 1);
          LongWord Identifier      = 0;
          if ((b & 0x1F) == 0x1F)
            {
              for (uint i = 0;;++i)
                {
                  if ((p >= p1) || (i >= 4))
                    {
                      Result = False;
                      CertErrorCode = cec_IdDataToBig;
                      break;
                    }
                  b = *p++;
                  Identifier = (Identifier << 7) | (b & 0x7F);
                  if (!(b & 0x80))
                    break;
                }
            }
          else
            Identifier = (b & 0x1F);
          if (Result)
            {
              LongWord Length = 0;
              if (p >= p1)
                {
                  Result = False;
                  CertErrorCode = cec_NoLengthData;
                }
              else
                {
                  b = *p++;
                  if (b & 0x80)
                    {
                      Byte LengthSize = (Byte)(b & 0x7F);
                      if (!LengthSize)
                        {
                          Result = False;
                          CertErrorCode = cec_IndefiniteFormNotSupported;
                        }
                      else if (LengthSize > 4)
                        {
                          Result = False;     
                          CertErrorCode = cec_LengthDataTooBig;
                        }
                      else
                        {
                          for (uint i = 0;i < LengthSize;++i)
                            {
                              if (p >= p1)
                                {
                                  Result = False;
                                  CertErrorCode = cec_InsufficientLengthData;
                                  break;
                                }
                              b = *p++;
                              Length = (Length << 8) | b;
                            }
                        }
                    }
                  else
                    Length = (b & 0x7F);
                  if (Result)
                    {
                      if (Length > Size)
                        {
                          Result = False;
                          CertErrorCode = cec_InvalidLengthData;
                        }
                      else
                        {
                          PByte ContentEnd = p + Length;
                          if (ContentEnd > p1)
                            {
                              Result = False;
                              CertErrorCode = cec_InvalidContentData;
                            }
                          else
                            {
                              if (IdentifierPC)
                                {
                                  Result = DoProcess(Depth + 1,p,Length,pStart,pEnd);
                                  if (Result)
                                    p += Length;
                                }
                              else
                                {
                                  Result = ProcessEntry(Identifier,IdentifierClass,p,Length);
                                  if (!Result)
                                    CertErrorCode = cec_ProcessEntryError;
                                  p += Length;
                                }
                            }
                        }
                    }
                }
            }
          if (!Result)
            break;
        }
    }
  return Result;
}

Boolean TProcessCertificate::Process(Pointer Data,uint DataSize)
{
  Boolean Result = False;
  if (!(Data && DataSize))
    CertErrorCode = cec_BadInfo;
  else
    {
      PByte p    = (PByte)Data;
      PByte pEnd = p + DataSize;
      Result = DoProcess(0,p,DataSize,p,pEnd);
      if (Result)
        CertErrorCode = cec_Ok;
    }
  return Result;
}

// TMyProcessCertificate ///////////////////////////////////////////////////////

class TMyProcessCertificate : public TProcessCertificate
  {
  public:
    uint CertFormatState;
    TMyProcessCertificate(void);
    virtual Boolean ProcessEntry(LongWord Id,Byte Class,PByte Data,uint DataSize);
    Boolean GetModulus(PByte *Data,Puint DataSize);
    PChar   GetSH1Digest(void);
    PChar   GetCertificateCN(void);
  private:
    uint  f;
    PByte Modulus;
    uint  ModulusSize;
    Char  SH1Digest[80];
    Char  CertificateCN[256];
  };
typedef TMyProcessCertificate * PMyProcessCertificate;

TMyProcessCertificate::TMyProcessCertificate(void)
{
  CertFormatState  = 0;
  f                = 0;
  Modulus          = NULL;
  ModulusSize      = 0;
  SH1Digest[0]     = '\0';
  CertificateCN[0] = '\0';
}

Boolean TMyProcessCertificate::GetModulus(PByte *Data,Puint DataSize)
{
  Boolean Result = False;
  if (Modulus)
    {
      *Data     = Modulus;
      *DataSize = ModulusSize;
      Result = True;
    }
  else
    {
      *Data     = NULL;
      *DataSize = 0;
    }
  return Result;
}

PChar TMyProcessCertificate::GetSH1Digest(void)
{
  return SH1Digest;
}

PChar TMyProcessCertificate::GetCertificateCN(void)
{
  return CertificateCN;
}

Boolean TMyProcessCertificate::ProcessEntry(LongWord Id,Byte Class,PByte Data,uint DataSize)
{
  Boolean Result = True;
  if (Id == 6)
    {
      if ((DataSize == 9) && !memcmp(Data,"\x2A\x86\x48\x86\xF7\x0D\x01\x01\x01",DataSize))
        {
          f = 1;
          CertFormatState = 1;
        }
      else if ((DataSize == 3) && !memcmp(Data,"\x55\x04\x03",DataSize))
        {
          f = 2;
        }
      else
        f = 0;
    }
  else if ((Id == 3) && (Class == 0) && (f == 1))
    {
      CertFormatState = 2;
      Result = False;
      PByte p    = Data;
      uint  Size = DataSize;
      if (!((Size > 0) && !p[0]))
        CertFormatState = 3;
      else
        {
          ++p;
          --Size;
          if (!((Size > 0) && (p[0] == 0x30)))
            CertFormatState = 4;
          else
            {
              ++p;
              --Size;

              if (!(Size > 0))
                CertFormatState = 5;
              else
                {
                  Byte b = *p++;
                  --Size;

                  Boolean IsOk = False;
                  uint Size2 = 0;
                  if (!(b & 0x80))
                    {
                      Size2 = (b & 0x7F);
                      IsOk = True;
                    }
                  else if (b == 0x81)
                    {
                      if (Size > 0)
                        {
                          Size2 = *p++;
                          --Size;
                          IsOk = True;
                        }
                    }
                  else if (b == 0x82)
                    {
                      if (Size > 1)
                        {
                          Size2 = (((uint)p[0]) << 8) | p[1];
                          p    += 2;
                          Size -= 2;
                          IsOk = True;
                        }
                    }
                  if (!IsOk)
                    CertFormatState = 6;
                  else
                    {
                      if (!((Size > 0) && (Size2 > 0) && (*p == 0x02)))
                        CertFormatState = 7;
                      else
                        {
                          ++p;
                          --Size;
                          --Size2;

                          IsOk = False;
                          if (!((Size > 0) && (Size2 > 0)))
                            CertFormatState = 8;
                          else
                            {
                              Byte b = *p++;
                              --Size;
                              --Size2;

                              uint Size3 = 0;
                              if (!(b & 0x80))
                                {
                                  Size3 = (b & 0x7F);
                                  IsOk = True;
                                }
                              else if (b == 0x81)
                                {
                                  if ((Size > 0) && (Size2 > 0))
                                    {
                                      Size3 = *p++;
                                      --Size;
                                      --Size2;
                                      if ((Size3 < Size2) && (Size3 < Size))
                                        IsOk = True;
                                    }
                                }
                              else if (b == 0x82)
                                {
                                  if ((Size > 1) && (Size2 > 1))
                                    {
                                      Size3  = (((uint)p[0]) << 8) | p[1];
                                      p     += 2;
                                      Size  -= 2;
                                      Size2 -= 2;
                                      if ((Size3 < Size2) && (Size3 < Size))
                                        IsOk = True;
                                    }
                                }
                              if (!IsOk)
                                CertFormatState = 9;
                              else
                                {
                                  if ((Size3 > 0) && !p[0])
                                    {
                                      ++p;
                                      --Size3;
                                    }

                                  Modulus     = p;
                                  ModulusSize = Size3;

                                  Char tmp[4 * 1024],tmp2[80];
                                  StringCbCopy((Pchar)tmp,sizeof tmp,(Pchar)"Modulus=");
                                  for (uint i = 0;i < ModulusSize;++i)
                                    {
                                      StringCbPrintf((Pchar)tmp2,sizeof tmp2,(Pchar)"%02X",Modulus[i]);
                                      StringCbCat((Pchar)tmp,sizeof tmp,(Pchar)tmp2);
                                    }
                                  StringCbCat((Pchar)tmp,sizeof tmp,(Pchar)"\x0A");

                                  PSHA1 S = new TSHA1();
                                  if (!S)
                                    CertFormatState = 10;
                                  else
                                    {
                                      if (!S->Update(tmp,(uint)strlen((Pchar)tmp)))
                                        CertFormatState = 11;
                                      else
                                        {
                                          if (!S->Final())
                                            CertFormatState = 12;
                                          else
                                            {
                                              PChar s = S->GetDigestStr();
                                              if (s && s[0])
                                                {
                                                  StringCbCopy((Pchar)SH1Digest,sizeof SH1Digest,(Pchar)s);
                                                  Result = True;
                                                }
                                              else
                                                CertFormatState = 13;
                                            }
                                        }
                                      delete S;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
  else if (((Id == 12) || (Id == 19) || (Id == 20)) && (Class == 0) && (f == 2))
    {
      uint l = DataSize;
      if (l > (sizeof CertificateCN - 1))
        l = (sizeof CertificateCN - 1);
      memcpy(CertificateCN,Data,l);
      CertificateCN[l] = '\0';
    }
  return Result;
}

////////////////////////////////////////////////////////////////////////////////

typedef struct
  {
    PChar dpp_CN;
    PByte dpp_CertificateData;
    uint  dpp_CertificateSize;
  } TDlgProcParam,* PDlgProcParam;

static INT_PTR CALLBACK DlgProc(HWND hdlg,UINT wMsg,WPARAM wParam,LPARAM lParam)
{
  Char tmp[1024],tmp1[256];
  PDlgProcParam Param;
  switch(wMsg)
    {
      case WM_INITDIALOG:
        Param = (PDlgProcParam)lParam;
        if (Param)
          {
#if defined(OS_WIN64)
            SetWindowLongPtr(hdlg,DWLP_USER,(LONG_PTR)Param);
#else
            SetWindowLong(hdlg,DWL_USER,(LONG)Param);
#endif
            CenterDialog(DllInst,hdlg);
            GetModuleFileName(NULL,(Pchar)tmp,sizeof tmp);
            SendDlgItemMessage(hdlg,100,WM_SETTEXT,0,(LPARAM)tmp);
            SendDlgItemMessage(hdlg,101,WM_SETTEXT,0,(LPARAM)((Param->dpp_CN && Param->dpp_CN[0]) ? Param->dpp_CN : (PChar)"???"));
            SetFocus(GetDlgItem(hdlg,2));
          }
        return FALSE;

      case WM_CLOSE:
        EndDialog(hdlg,FALSE);
        return TRUE;

      case WM_COMMAND:
        switch(wParam)
          {
            case 1:
              EndDialog(hdlg,TRUE);
              return TRUE;

            case 2:
              EndDialog(hdlg,FALSE);
              return TRUE;

            case 8:
#if defined(OS_WIN64)
              Param = (PDlgProcParam)GetWindowLongPtr(hdlg,DWLP_USER);
#else
              Param = (PDlgProcParam)GetWindowLong(hdlg,DWL_USER);
#endif
              if (Param && Param->dpp_CertificateData)
                {
                  LongWord NativeErrorCode = 0;
                  const PChar   ServerName = (PChar)"www.heise.de";
                  const uint    ServerPort = 0;
                  const PChar   URL        = (PChar)"/netze/tools/certainty";
                  const Boolean Secure     = False;
                  if (HTTPRequest(ServerName,ServerPort,URL,Secure,Param->dpp_CertificateData,Param->dpp_CertificateSize,&NativeErrorCode,L"Content-Type: application/x-x509-ca-cert",NULL))
                    {
                      LoadString(DllInst,LangResOfs + 1,(Pchar)tmp1,sizeof tmp1);
                      StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)tmp1,Param->dpp_CN,"Heise"/*ServerName*/,""/*URL*/);
                    }
                  else
                    {
                      LoadString(DllInst,LangResOfs + 2,(Pchar)tmp1,sizeof tmp1);
                      StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)tmp1,Param->dpp_CN,"Heise"/*ServerName*/,""/*URL*/,NativeErrorCode);
                    }
                  MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_TOPMOST | MB_SETFOREGROUND | MB_OK);
                }
              return TRUE;

            case 9:
              {
                SHELLEXECUTEINFOA SEI;
                memset(&SEI,0,sizeof SEI);
                SEI.cbSize   = sizeof SEI;
                SEI.fMask   |= 0x00800000/*SEE_MASK_NOZONECHECKS*/;
                SEI.fMask   |= SEE_MASK_FLAG_DDEWAIT;
                SEI.lpVerb   = "open";
                SEI.nShow    = SW_SHOWNORMAL;
                if (LangResOfs == 1000)
                  SEI.lpFile = (Pchar)"http://www.heise.de/ct/projekte/sslwaechter/";
                else
                  SEI.lpFile = (Pchar)"http://www.heise-online.co.uk/security/features/111039";
                ShellExecuteEx(&SEI);
              }
              return TRUE;
          }
        break;
    }
  return FALSE;
}

#if defined(OS_WIN32)
#pragma comment(linker, "/EXPORT:CertDllVerifyRevocation=_CertDllVerifyRevocation@28")
extern "C"
#elif defined(OS_WIN64)
extern "C" __declspec(dllexport)
#endif
BOOL WINAPI CertDllVerifyRevocation(/*__in*/ DWORD dwEncodingType,/*__in*/ DWORD dwRevType,/*__in*/ DWORD cContext,/*__in_ecount(cContext)*/ PVOID rgpvContext[],/*__in*/ DWORD /*dwFlags*/,/*__in_opt*/ PCERT_REVOCATION_PARA /*pRevPara*/,/*__inout*/ PCERT_REVOCATION_STATUS pRevStatus)
{
  BOOL Result = FALSE;
  if (!cContext)
    SetLastError((DWORD)E_INVALIDARG);
  else
    {
      Char tmp[256],tmp1[256];
      Boolean CheckError = False;
      Boolean DoRevoke   = False;

      if (dwEncodingType != X509_ASN_ENCODING)
        CheckError = True;
      else if (dwRevType != CERT_CONTEXT_REVOCATION_TYPE)
        CheckError = True;
      else if (!rgpvContext)
        CheckError = True;
      else
        {
          PCCERT_CONTEXT *C = (PCCERT_CONTEXT *)rgpvContext;
          for (uint i = 0;!CheckError && (i < cContext);++i)
            {
              if (!C[i] || !C[i]->pbCertEncoded || !C[i]->cbCertEncoded)
                CheckError = True;
              else
                {
                  Char SH1Digest[80];
                  Char CertificateCN[256];
                  SH1Digest[0]     = '\0';
                  CertificateCN[0] = '\0';
                  TProcessCertificate::TCertErrorCode CertErrorCode   = TProcessCertificate::cec_UnknownError;
                  uint                                CertFormatState = 0;
                  PMyProcessCertificate PC = new TMyProcessCertificate();
                  if (!PC)
                    {
                      CheckError = True;
                      break;
                    }

                  if (!PC->Process(C[i]->pbCertEncoded,C[i]->cbCertEncoded))
                    {
                      CheckError = True;
                      CertErrorCode   = PC->CertErrorCode;
                      CertFormatState = PC->CertFormatState;
                    }
                  else
                    {
                      PChar s = PC->GetSH1Digest();
                      if (s)
                        {
                          uint l = (uint)strlen((Pchar)s);
                          if (l > 20)
                            s += 20;
                          StringCbCopy((Pchar)SH1Digest,sizeof SH1Digest,(Pchar)s);
                        }
                      s = PC->GetCertificateCN();
                      if (s)
                        StringCbCopy((Pchar)CertificateCN,sizeof CertificateCN,(Pchar)s);
                    }
                  delete PC;

                  if (!SH1Digest[0])
                    {
                      LoadString(DllInst,LangResOfs + 3,(Pchar)tmp1,sizeof tmp1);
                      StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)tmp1,CertErrorCode,CertFormatState);
                      MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_TOPMOST | MB_SETFOREGROUND | MB_OK);
                    }
                  else
                    {
                      Boolean FoundInCache = False;
                      EnterCriticalSection(&SHA1CheckCacheCS);
                      if (SHA1CheckCache)
                        {
                          for (uint i = 0;i < SHA1CheckCacheCount;++i)
                            if (!strcmp((Pchar)SH1Digest,(Pchar)SHA1CheckCache[i].SHA1))
                              {
                                DoRevoke = SHA1CheckCache[i].CheckResult;
                                FoundInCache = True;
                                break;
                              }
                        }
                      LeaveCriticalSection(&SHA1CheckCacheCS);

                      if (!FoundInCache)
                        {
                          uint ErrorCode1 = 0,ErrorCode2 = 0;
                          Char InstallDir[1024];
                          InstallDir[0] = '\0';
                          if (GetEnvironmentVariable("ProgramFiles",(Pchar)InstallDir,sizeof InstallDir))
                            {
                              StringCbCat((Pchar)InstallDir,sizeof InstallDir,AppInstallPath);
                              StringCbCat((Pchar)InstallDir,sizeof InstallDir,"\\");
                            }
                          if (!InstallDir[0])
                            {
                              ErrorCode1 = 1;
                              ErrorCode2 = GetLastError();
                            }
                          else
                            {
                              Char InstallDir2[1024];
                              StringCbCopy((Pchar)InstallDir2,sizeof InstallDir2,(Pchar)InstallDir);
                              StringCbCat((Pchar)InstallDir2,sizeof InstallDir2,(Pchar)"*");
                              WIN32_FIND_DATA FindData;
                              HANDLE FindHandle = FindFirstFile((Pchar)InstallDir2,&FindData);
                              if (FindHandle == INVALID_HANDLE_VALUE)
                                {
                                  ErrorCode1 = 2;
                                  ErrorCode2 = GetLastError();
                                }
                              else
                                {
                                  do
                                    {
                                      if (!(FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                                        {
                                          PChar Name = (PChar)FindData.cFileName;
                                          if (((Char)CharUpper((LPTSTR)Name[0]) == 'B') && ((Char)CharUpper((LPTSTR)Name[1]) == 'L') && ((Char)CharUpper((LPTSTR)Name[2]) == 'A') && ((Char)CharUpper((LPTSTR)Name[3]) == 'C') && ((Char)CharUpper((LPTSTR)Name[4]) == 'K') && ((Char)CharUpper((LPTSTR)Name[5]) == 'L') && ((Char)CharUpper((LPTSTR)Name[6]) == 'I') && ((Char)CharUpper((LPTSTR)Name[7]) == 'S') && ((Char)CharUpper((LPTSTR)Name[8]) == 'T'))
                                            {
                                              Char FileName[1024];
                                              StringCbCopy((Pchar)FileName,sizeof FileName,(Pchar)InstallDir);
                                              StringCbCat((Pchar)FileName,sizeof FileName,(Pchar)Name);
                                              if (!FindData.nFileSizeHigh && (FindData.nFileSizeLow < 128 * 1024 * 1024))
                                                {
                                                  uint FSize = FindData.nFileSizeLow;
                                                  PByte Buf = new Byte[FSize + 1];
                                                  if (!Buf)
                                                    {
                                                      ErrorCode1 = 3;
                                                    }
                                                  else
                                                    {
                                                      Buf[FSize] = '\0';
                                                      HANDLE H = CreateFile((Pchar)FileName,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
                                                      if (H == INVALID_HANDLE_VALUE)
                                                        {
                                                          ErrorCode1 = 4;
                                                          ErrorCode2 = GetLastError();
                                                        }
                                                      else
                                                        {
                                                          DWORD R;
                                                          if (!ReadFile(H,Buf,FSize,&R,NULL) || (R != FSize))
                                                            {
                                                              ErrorCode1 = 5;
                                                              ErrorCode2 = GetLastError();
                                                            }
                                                          else
                                                            {
                                                              if (strstr((Pchar)Buf,(Pchar)SH1Digest))
                                                                {
                                                                  StringCbPrintf((Pchar)tmp,sizeof tmp,"Weak certificate found: '%s'",CertificateCN);
                                                                  HANDLE EL = RegisterEventSource(NULL,AppName);
                                                                  if (EL)
                                                                    {
                                                                      LPCTSTR S[1] = {(LPCTSTR)tmp};
                                                                      ReportEvent(EL,EVENTLOG_WARNING_TYPE,0,EVMSG_GENERIC,NULL,sizeof S / sizeof S[0],0,S,0);
                                                                      DeregisterEventSource(EL);
                                                                    }

                                                                  if (!IsInteractive)
                                                                    DoRevoke = True;
                                                                  else
                                                                    {
                                                                      TDlgProcParam DPP;
                                                                      memset(&DPP,0,sizeof DPP);
                                                                      DPP.dpp_CN              = CertificateCN;
                                                                      DPP.dpp_CertificateData = C[i]->pbCertEncoded;
                                                                      DPP.dpp_CertificateSize = C[i]->cbCertEncoded;
                                                                      if (DialogBoxParam(DllInst,MAKEINTRESOURCE(LangResOfs + 100),NULL,DlgProc,(LPARAM)&DPP) != TRUE)
                                                                        {
                                                                          DoRevoke = True;
                                                                        }
                                                                    }
                                                                }
                                                            }
                                                          CloseHandle(H);
                                                        }
                                                      delete [] Buf;
                                                    }
                                                }
                                              if (DoRevoke)
                                                break;
                                            }
                                        }
                                    }
                                  while (FindNextFile(FindHandle,&FindData));
                                  FindClose(FindHandle);
                                }
                            }
                          if (ErrorCode1)
                            {
                              LoadString(DllInst,LangResOfs + 3,(Pchar)tmp1,sizeof tmp1);
                              StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)tmp1,ErrorCode1 + 100,ErrorCode2);
                              MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_TOPMOST | MB_SETFOREGROUND | MB_OK);
                            }

                          if (SHA1CheckCache)
                            {
                              EnterCriticalSection(&SHA1CheckCacheCS);
                              if (SHA1CheckCacheCount < SHA1CheckCacheSize)
                                {
                                  StringCbCopy((Pchar)SHA1CheckCache[SHA1CheckCacheCount].SHA1,sizeof SHA1CheckCache[SHA1CheckCacheCount].SHA1,(Pchar)SH1Digest);
                                  SHA1CheckCache[SHA1CheckCacheCount].CheckResult = DoRevoke;
                                  ++SHA1CheckCacheCount;
                                }
                              LeaveCriticalSection(&SHA1CheckCacheCS);
                            }
                        }
                    }
                }
            }
        }
      if (CheckError)
        {
          Result = FALSE;
          if (pRevStatus)
            pRevStatus->dwError = (DWORD)CRYPT_E_NO_REVOCATION_CHECK;
        }
      else if (!DoRevoke)
        {
          Result = FALSE;
          if (pRevStatus)
            pRevStatus->dwError = (DWORD)CRYPT_E_NO_REVOCATION_CHECK;
        }
      else
        {
          Result = FALSE;
          if (pRevStatus)
            pRevStatus->dwError = (DWORD)CRYPT_E_REVOKED;
        }
    }
  return Result;
}

// DllMain /////////////////////////////////////////////////////////////////////

BOOL WINAPI DllMain(HINSTANCE hinstDLL,DWORD fdwReason,LPVOID /*lpReserved*/)
{
  BOOL Result = TRUE;
  switch(fdwReason)
    {
      case DLL_PROCESS_ATTACH:
        DllInst = hinstDLL;
        if (!Init())
          Result = FALSE;
        break;

      case DLL_PROCESS_DETACH:
        if (!Done())
          Result = FALSE;
        break;
    }
  return Result;
}

