////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// setup.cpp                                                                  //
// (c) 2008 Matthias Withopf / Heise Zeitschriften Verlag                     //
//   http://www.heise.de/ct/projekte/sslwaechter/                             //
//   http://www.withopf.com/tools/sslguardian/                                //
//                                                                            //
//  This file is part of SSLGuardian.                                         //
//                                                                            //
//  SSLGuardian is free software: you can redistribute it and/or modify       //
//  it under the terms of the GNU General Public License as published by      //
//  the Free Software Foundation, either version 3 of the License, or         //
//  (at your option) any later version.                                       //
//                                                                            //
//  SSLGuardian is distributed in the hope that it will be useful,            //
//  but WITHOUT ANY WARRANTY; without even the implied warranty of            //
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             //
//  GNU General Public License for more details.                              //
//                                                                            //
//  You should have received a copy of the GNU General Public License         //
//  along with SSLGuardian.  If not, see <http://www.gnu.org/licenses/>.      //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

#include "SSLGuardian.h"

#if defined(CPU_AMD64)
const PChar SignonStr = (PChar)"Setup SSLGuardian V1.1 [x64] (c) 2008 Matthias Withopf / Heise Zeitschriften Verlag";
#elif defined(CPU_IA64)
const PChar SignonStr = (PChar)"Setup SSLGuardian V1.1 [IA-64] (c) 2008 Matthias Withopf / Heise Zeitschriften Verlag";
#else
const PChar SignonStr = (PChar)"Setup SSLGuardian V1.1 (c) 2008 Matthias Withopf / Heise Zeitschriften Verlag";
#endif

#define STRICT
#define WIN32_LEAN_AND_MEAN
#pragma warning(disable:4201)
#include <windows.h>
#include <wincrypt.h>
#include <shellapi.h>
#include <strsafe.h>
#include <aclapi.h>
#pragma warning(default:4201)
#include "utils.h"

HINSTANCE AppInst    = NULL;
LPSTR     AppCmdLine = NULL;

#if defined(CPU_INTEL)
const PChar SSLGuardianDLLFileName = (PChar)"SSLGuardian.dll";
#elif defined(CPU_AMD64)
const PChar SSLGuardianDLLFileName = (PChar)"SSLGuardian-x64.dll";
#elif defined(CPU_IA64)
const PChar SSLGuardianDLLFileName = (PChar)"SSLGuardian-ia64.dll";
#endif

static HKEY CreateAppKey(void)
{
  HKEY Result = NULL;
  SID_IDENTIFIER_AUTHORITY SIDAuthWorld = SECURITY_WORLD_SID_AUTHORITY;
  PSID EveryoneSID = NULL;
  if (AllocateAndInitializeSid(&SIDAuthWorld,1,SECURITY_WORLD_RID,0,0,0,0,0,0,0,&EveryoneSID))
    {
      EXPLICIT_ACCESS EA;
      memset(&EA,0,sizeof EA);
      EA.grfAccessPermissions = KEY_ALL_ACCESS; 
      EA.grfAccessMode        = SET_ACCESS;
      EA.grfInheritance       = NO_INHERITANCE;
      EA.Trustee.TrusteeForm  = TRUSTEE_IS_SID;
      EA.Trustee.TrusteeType  = TRUSTEE_IS_WELL_KNOWN_GROUP;
      EA.Trustee.ptstrName    = (LPTSTR)EveryoneSID;

      PACL ACL = NULL;
      DWORD Res = SetEntriesInAcl(1,&EA,NULL,&ACL);
      if (Res == ERROR_SUCCESS)
        {
          PSECURITY_DESCRIPTOR SD = (PSECURITY_DESCRIPTOR)new Byte[SECURITY_DESCRIPTOR_MIN_LENGTH];
          if (SD)
            {
              if (InitializeSecurityDescriptor(SD,SECURITY_DESCRIPTOR_REVISION))
                {
                  if (SetSecurityDescriptorDacl(SD,TRUE,ACL,FALSE))
                    {
                      SECURITY_ATTRIBUTES SA;
                      SA.nLength              = sizeof SA;
                      SA.lpSecurityDescriptor = SD;
                      SA.bInheritHandle       = FALSE;
                      DWORD Disposition;
                      HKEY  SubKey = NULL;
                      LONG  Res2 = RegCreateKeyEx(HKEY_LOCAL_MACHINE,AppRegKey,0,"",0,KEY_READ | KEY_WRITE,&SA,&SubKey,&Disposition);
                      if (Res2 == ERROR_SUCCESS)
                        {
                          Result = SubKey;
                        }
                    }
                }
              delete [] SD;
            }
          LocalFree(ACL);
        }
      FreeSid(EveryoneSID);
    }
  return Result;
}

static Boolean DoActivate2(uint Mode,PChar AKeyName)
{
  Boolean Result = False;
  Char KeyName[256];
  KeyName[0] = '\0';
  if (AKeyName)
    {
      StringCbCopy((Pchar)KeyName,sizeof KeyName,(Pchar)AKeyName);
      StringCbCat((Pchar)KeyName,sizeof KeyName,"\\");
    }
  StringCbCat((Pchar)KeyName,sizeof KeyName,"Software\\Microsoft\\Windows\\CurrentVersion\\Internet Settings");
  HKEY Key2;
  DWORD dw;
  LONG L = RegCreateKeyEx(!Mode ? HKEY_USERS : HKEY_CURRENT_USER,(Pchar)KeyName,0,0,REG_OPTION_NON_VOLATILE,KEY_SET_VALUE,NULL,&Key2,&dw);
  if (L == ERROR_SUCCESS)
    {
      DWORD CurValue = 1;
      RegSetValueEx(Key2,"CertificateRevocation",0,REG_DWORD,(BYTE *)&CurValue,sizeof CurValue);
      RegCloseKey(Key2);
      Result = True;
    }
  return Result;
}

static Boolean DoActivate(uint Mode)
{
  Boolean Result = False;
  if (Mode)
    Result = DoActivate2(Mode,NULL);
  else
    {
      {
        HKEY Key;
        LONG L = RegOpenKeyEx(HKEY_USERS,"",0,KEY_ENUMERATE_SUB_KEYS,&Key);
        if (L == ERROR_SUCCESS)
          {
            Result = True;
            for (uint CurIndex = 0;;++CurIndex)
              {
                Char  KeyName[256];
                DWORD KeyNameSize = sizeof KeyName;
                LONG L = RegEnumKeyEx(Key,CurIndex,(Pchar)KeyName,&KeyNameSize,NULL,NULL,NULL,NULL);
                if (L != ERROR_SUCCESS)
                  break;
                if (!strstr((Pchar)KeyName,"_Classes"))
                  {
                    if (!DoActivate2(Mode,KeyName))
                      {
                        Result = False;
                        break;
                      }
                  }
              }
            RegCloseKey(Key);
          }
      }
      {
        Boolean IsOk = False;
        const uint MaxValueSize = 1 * 1024;
        UCHAR tmp[MaxValueSize];
        UCHAR ProfileDir[MaxValueSize];
        ProfileDir[0] = '\0';
        PChar KeyName    = (PChar)"SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList";
        PChar ValueName  = (PChar)"ProfilesDirectory";
        HKEY Key;
        LONG L = RegOpenKeyEx(HKEY_LOCAL_MACHINE,(Pchar)KeyName,0,KEY_QUERY_VALUE,&Key);
        if (L == ERROR_SUCCESS)
          {
            memset(tmp,0,sizeof tmp);
            DWORD Type;
            DWORD Size = MaxValueSize;
            L = RegQueryValueEx(Key,(Pchar)ValueName,0,&Type,(BYTE *)tmp,&Size);
            if ((L == ERROR_SUCCESS) && (Type == REG_EXPAND_SZ))
              {
                if (ExpandEnvironmentStrings((LPCSTR)tmp,(LPSTR)ProfileDir,MaxValueSize))
                  IsOk = True;
              }
            RegCloseKey(Key);
          }
        if (IsOk && ProfileDir[0])
          {
            HANDLE Token = NULL;
            if (OpenProcessToken(GetCurrentProcess(),TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY,&Token))
              {
                LUID PrivNameValue;
                if (LookupPrivilegeValue(NULL,SE_BACKUP_NAME,&PrivNameValue))
                  {
                    TOKEN_PRIVILEGES Privileges;
                    Privileges.PrivilegeCount           = 1;
                    Privileges.Privileges[0].Luid       = PrivNameValue;
                    Privileges.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
                    DWORD dwRet;
                    if (!AdjustTokenPrivileges(Token,FALSE,&Privileges,sizeof(Privileges),NULL,&dwRet) || GetLastError() == ERROR_NOT_ALL_ASSIGNED)
                      {
                      }
                  }
                if (LookupPrivilegeValue(NULL,SE_RESTORE_NAME,&PrivNameValue))
                  {
                    TOKEN_PRIVILEGES Privileges;
                    Privileges.PrivilegeCount           = 1;
                    Privileges.Privileges[0].Luid       = PrivNameValue;
                    Privileges.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
                    DWORD dwRet;
                    if (!AdjustTokenPrivileges(Token,FALSE,&Privileges,sizeof(Privileges),NULL,&dwRet) || GetLastError() == ERROR_NOT_ALL_ASSIGNED)
                      {
                      }
                  }
                CloseHandle(Token);
              }

            StringCbCopy((Pchar)tmp,sizeof tmp,(Pchar)ProfileDir);
            StringCbCat((Pchar)tmp,sizeof tmp,(Pchar)"\\*");
            WIN32_FIND_DATA FindData;
            HANDLE FindHandle = FindFirstFile((Pchar)tmp,&FindData);
            if (FindHandle != INVALID_HANDLE_VALUE)
              {
                do
                  {
                    if (FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
                      {
                        PChar Name = (PChar)FindData.cFileName;
                        if (!(((Name[0] == '.') && !Name[1]) || ((Name[0] == '.') && (Name[1] == '.') && !Name[2])))
                          {
                            StringCbCopy((Pchar)tmp,sizeof tmp,(Pchar)ProfileDir);
                            StringCbCat((Pchar)tmp,sizeof tmp,(Pchar)"\\");
                            StringCbCat((Pchar)tmp,sizeof tmp,(Pchar)Name);
                            StringCbCat((Pchar)tmp,sizeof tmp,(Pchar)"\\ntuser.dat");
                            DWORD A = GetFileAttributes((Pchar)tmp);
                            if ((A != INVALID_FILE_ATTRIBUTES) && !(A & FILE_ATTRIBUTE_DIRECTORY))
                              {
                                PChar LoadedUserName = (PChar)"LoadedUser";
                                LONG R = RegLoadKey(HKEY_USERS,(Pchar)LoadedUserName,(Pchar)tmp);
                                if (R == ERROR_SUCCESS)
                                  {
                                    DoActivate2(Mode,LoadedUserName);
                                    RegUnLoadKey(HKEY_USERS,(Pchar)LoadedUserName);
                                  }
                              }
                          }
                      }
                  }
                while (FindNextFile(FindHandle,&FindData));
                FindClose(FindHandle);
              }
          }
      }
    }
  return Result;
}

static Boolean IsInstalled(void)
{
  Boolean Result = False;
  PChar KeyName    = (PChar)"SOFTWARE\\Microsoft\\Cryptography\\OID\\EncodingType 1\\CertDllVerifyRevocation\\DEFAULT";
  PChar ValueName  = (PChar)"Dll";
  HKEY Key;
  LONG L = RegOpenKeyEx(HKEY_LOCAL_MACHINE,(Pchar)KeyName,0,KEY_QUERY_VALUE,&Key);
  if (L == ERROR_SUCCESS)
    {
      const uint MaxValueSize = 8 * 1024;
      BYTE Value[MaxValueSize];
      memset(Value,0,sizeof Value);
      DWORD Type;
      DWORD Size = MaxValueSize;
      L = RegQueryValueEx(Key,(Pchar)ValueName,0,&Type,(BYTE *)Value,&Size);
      if ((L == ERROR_SUCCESS) && (Type == REG_MULTI_SZ))
        {
          for (PChar p = Value;p[0] && p[1];)
            {
              PChar p1 = (PChar)strrchr((Pchar)p,'\\');
              if (p1)
                {
                  ++p1;
                  if (!_strcmpi((Pchar)p1,(Pchar)SSLGuardianDLLFileName))
                    {
                      Result = True;
                      break;
                    }
                }
              uint l = (uint)strlen((Pchar)p);
              p += (l + 1);
            }
        }
      RegCloseKey(Key);
    }
  return Result;
}

static Boolean ChangeCRLDownload(Boolean Change)
{
  Boolean Result = False;
  PChar KeyName    = (PChar)"SOFTWARE\\Microsoft\\Cryptography\\OID\\EncodingType 1\\CertDllVerifyRevocation\\DEFAULT";
  PChar ValueName  = (PChar)"Dll";
  PChar ValueName2 = (PChar)"DllOrig";
  HKEY Key;
  LONG L = RegOpenKeyEx(HKEY_LOCAL_MACHINE,(Pchar)KeyName,0,KEY_QUERY_VALUE | KEY_SET_VALUE,&Key);
  if (L == ERROR_SUCCESS)
    {
      const uint MaxValueSize = 8 * 1024;
      BYTE Value[MaxValueSize];
      if (Change)
        {
          DWORD Type;
          DWORD Size = MaxValueSize;
          L = RegQueryValueEx(Key,(Pchar)ValueName,0,&Type,(BYTE *)Value,&Size);
          if ((L == ERROR_SUCCESS) && (Type == REG_MULTI_SZ))
            {
              L = RegSetValueEx(Key,(Pchar)ValueName2,0,Type,Value,Size);
              if (L == ERROR_SUCCESS)
                {
                  const uint CheckSize = 12;
                  if (Size >= CheckSize)
                    {
                      for (uint i = 0;i < Size - CheckSize;++i)
                        {
                          if (((Char)CharUpper((LPTSTR)Value[i + 0]) == 'C') && ((Char)CharUpper((LPTSTR)Value[i + 1]) == 'R') && ((Char)CharUpper((LPTSTR)Value[i + 2]) == 'Y') && ((Char)CharUpper((LPTSTR)Value[i + 3]) == 'P') && ((Char)CharUpper((LPTSTR)Value[i + 4]) == 'T')
                           && ((Char)CharUpper((LPTSTR)Value[i + 5]) == 'N') && ((Char)CharUpper((LPTSTR)Value[i + 6]) == 'E') && ((Char)CharUpper((LPTSTR)Value[i + 7]) == 'T') && ((Char)CharUpper((LPTSTR)Value[i + 8]) == '.')
                           && ((Char)CharUpper((LPTSTR)Value[i + 9]) == 'D') && ((Char)CharUpper((LPTSTR)Value[i + 10]) == 'L') && ((Char)CharUpper((LPTSTR)Value[i + 11]) == 'L') && !Value[i + 12])
                            {
                              memmove(Value + i,Value + i + CheckSize,Size - (i + CheckSize));
                              Size -= CheckSize;
                              break;
                            }
                        }
                      L = RegSetValueEx(Key,(Pchar)ValueName,0,Type,Value,Size);
                      if (L == ERROR_SUCCESS)
                        Result = True;
                    }
                }
            }
        }
      else
        {
          DWORD Type;
          DWORD Size = MaxValueSize;
          L = RegQueryValueEx(Key,(Pchar)ValueName2,0,&Type,(BYTE *)Value,&Size);
          if ((L == ERROR_SUCCESS) && (Type == REG_MULTI_SZ))
            {
              L = RegSetValueEx(Key,(Pchar)ValueName,0,Type,Value,Size);
              if (L == ERROR_SUCCESS)
                Result = True;
              RegDeleteValue(Key,(Pchar)ValueName2);
            }
        }
      RegCloseKey(Key);
    }
  return Result;
}

static Boolean DoSetup2(uint ActivateMode,Boolean DisableCRLDownload,Boolean UseIEProxySettings,Boolean OnlineUpdate)
{
  Boolean Result = False;
  Char tmp[1024],tmp1[256];
  Boolean IsOk = False;
  SC_HANDLE hSC = OpenSCManager(NULL,NULL,GENERIC_READ | GENERIC_WRITE | GENERIC_EXECUTE);
  if (hSC)
    {
      IsOk = True;
      CloseServiceHandle(hSC);
    }
  if (!IsOk)
    {
      LoadString(AppInst,LangResOfs + 11,(Pchar)tmp,sizeof tmp);
      MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_OK);
    }
  else
    {
      Char InstallDir[1024];
      InstallDir[0] = '\0';
      if (GetEnvironmentVariable("ProgramFiles",(Pchar)InstallDir,sizeof InstallDir))
        {
          StringCbCat((Pchar)InstallDir,sizeof InstallDir,AppInstallPath);
          BOOL  f  = CreateDirectory((Pchar)InstallDir,NULL);
          DWORD LE = GetLastError();
          if (!f && (LE == 183))
            f = TRUE;
          if (!f)
            {
              LoadString(AppInst,LangResOfs + 1,(Pchar)tmp1,sizeof tmp1);
              StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)tmp1,InstallDir,LE);
              MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_OK);
            }
          else
            {
              GetModuleFileName(NULL,(Pchar)tmp,sizeof tmp);
              uint l = (uint)strlen((Pchar)tmp);
              if (l > 0)
                {
                  IsOk = False;
                  for (uint i = l;i > 0;--i)
                    if (tmp[i - 1] == '\\')
                      {
                        tmp[i] = '\0';
                        IsOk = True;
                        break;
                      }
                  if (IsOk)
                    {
                      Boolean IsInst = IsInstalled();

                      Char SSLGuardianFileName[1024];
                      SSLGuardianFileName[0] = '\0';
                      Char SrcDir[1024];
                      StringCbCopy((Pchar)SrcDir,sizeof SrcDir,(Pchar)tmp);
                      Char SrcDir2[1024];
                      StringCbCopy((Pchar)SrcDir2,sizeof SrcDir2,(Pchar)SrcDir);
                      StringCbCat((Pchar)SrcDir2,sizeof SrcDir2,(Pchar)"*");
                      WIN32_FIND_DATA FindData;
                      HANDLE FindHandle = FindFirstFile((Pchar)SrcDir2,&FindData);
                      if (FindHandle != INVALID_HANDLE_VALUE)
                        {
                          do
                            {
                              if (!(FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                                {
                                  Boolean DoCopy = False;
                                  Boolean IsDLL  = False;
                                  PChar Name = (PChar)FindData.cFileName;
                                  if (((Char)CharUpper((LPTSTR)Name[0]) == 'B') && ((Char)CharUpper((LPTSTR)Name[1]) == 'L') && ((Char)CharUpper((LPTSTR)Name[2]) == 'A') && ((Char)CharUpper((LPTSTR)Name[3]) == 'C') && ((Char)CharUpper((LPTSTR)Name[4]) == 'K') && ((Char)CharUpper((LPTSTR)Name[5]) == 'L') && ((Char)CharUpper((LPTSTR)Name[6]) == 'I') && ((Char)CharUpper((LPTSTR)Name[7]) == 'S') && ((Char)CharUpper((LPTSTR)Name[8]) == 'T'))
                                    DoCopy = True;
                                  else if (!_strcmpi((Pchar)Name,(Pchar)SSLGuardianDLLFileName))
                                    {
                                      DoCopy = True;
                                      IsDLL  = True;
                                    }
                                  if (DoCopy)
                                    {
                                      Char SrcFileName[1024];
                                      StringCbCopy((Pchar)SrcFileName,sizeof SrcFileName,(Pchar)SrcDir);
                                      StringCbCat((Pchar)SrcFileName,sizeof SrcFileName,(Pchar)Name);
                                      Char DestFileName[1024];
                                      StringCbCopy((Pchar)DestFileName,sizeof DestFileName,(Pchar)InstallDir);
                                      StringCbCat((Pchar)DestFileName,sizeof DestFileName,(Pchar)"\\");
                                      StringCbCat((Pchar)DestFileName,sizeof DestFileName,(Pchar)Name);
                                      if (IsDLL)
                                        StringCbCopy((Pchar)SSLGuardianFileName,sizeof SSLGuardianFileName,(Pchar)DestFileName);
                                      if (!IsInst)
                                        {
                                          if (!CopyFile((Pchar)SrcFileName,(Pchar)DestFileName,FALSE))
                                            {
                                              DWORD LE = GetLastError();
                                              LoadString(AppInst,LangResOfs + 2,(Pchar)tmp1,sizeof tmp1);
                                              StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)tmp1,SrcFileName,DestFileName,LE);
                                              MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_OK);
                                              IsOk = False;
                                              break;
                                            }
                                        }
                                    }
                                }
                            }
                          while (FindNextFile(FindHandle,&FindData));
                          FindClose(FindHandle);

                          if (IsOk && SSLGuardianFileName[0])
                            {
                              WCHAR FileName2[1024];
                              uint l = (uint)strlen((Pchar)SSLGuardianFileName);
                              if (l > (sizeof FileName2 / sizeof FileName2[0] - 1))
                                l = (sizeof FileName2 / sizeof FileName2[0] - 1);
                              uint i;
                              for (i = 0;i < l;++i)
                                FileName2[i] = SSLGuardianFileName[i];
                              FileName2[l] = '\0';

                              if (IsInst)
                                {
                                  ChangeCRLDownload(False);
                                  if (CryptUnregisterDefaultOIDFunction(X509_ASN_ENCODING,CRYPT_OID_VERIFY_REVOCATION_FUNC,FileName2))
                                    {
                                      LoadString(AppInst,LangResOfs + 5,(Pchar)tmp,sizeof tmp);
                                      MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_OK);
                                      Result = True;
                                    }
                                  else
                                    {
                                      LE = GetLastError();
                                      LoadString(AppInst,LangResOfs + 7,(Pchar)tmp1,sizeof tmp1);
                                      StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)tmp1,LE);
                                      MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_OK);
                                    }
                                }
                              else
                                {
                                  if (CryptRegisterDefaultOIDFunction(X509_ASN_ENCODING,CRYPT_OID_VERIFY_REVOCATION_FUNC,CRYPT_REGISTER_FIRST_INDEX,FileName2))
                                    {
                                      LoadString(AppInst,LangResOfs + 3,(Pchar)tmp,sizeof tmp);
                                      MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_OK);

                                      DoActivate(ActivateMode);

                                      if (DisableCRLDownload)
                                        ChangeCRLDownload(True);

                                      HKEY K = CreateAppKey();
                                      if (K)
                                        {
                                          DWORD CurValue = OnlineUpdate;
                                          RegSetValueEx(K,AppRegValue_CheckUpdate,0,REG_DWORD,(BYTE *)&CurValue,sizeof CurValue);
                                          RegCloseKey(K);
                                        }

                                      if (UseIEProxySettings)
                                        {
                                          STARTUPINFO SI;
                                          memset(&SI,0,sizeof SI);
                                          SI.cb = sizeof SI;
                                          PROCESS_INFORMATION PI;
                                          memset(&PI,0,sizeof PI);

                                          DWORD V = GetVersion();
                                          Byte  MajorVersion = (Byte)V;
                                          Char AppCmdLine[128];
                                          if (MajorVersion >= 6)
                                            StringCbCopy((Pchar)AppCmdLine,sizeof AppCmdLine,(Pchar)"netsh winhttp proxy import");
                                          else
                                            StringCbCopy((Pchar)AppCmdLine,sizeof AppCmdLine,(Pchar)"proxycfg -u");
                                          if (!CreateProcess(NULL,(Pchar)AppCmdLine,NULL,NULL,FALSE,0,NULL,NULL,&SI,&PI))
                                            {
                                              DWORD LE = GetLastError();
                                              LoadString(AppInst,LangResOfs + 13,(Pchar)tmp1,sizeof tmp1);
                                              StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)tmp1,LE);
                                              MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_OK);
                                            }
                                          if (PI.hProcess)
                                            CloseHandle(PI.hProcess);
                                          if (PI.hThread)
                                            CloseHandle(PI.hThread);
                                        }

                                      LoadString(AppInst,LangResOfs + 14,(Pchar)tmp,sizeof tmp);
                                      if (MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_YESNO) == IDYES)
                                        {
                                          PChar KeyName   = (PChar)"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\iexplore.exe";
                                          PChar ValueName = (PChar)"";
                                          HKEY Key;
                                          LONG L = RegOpenKeyEx(HKEY_LOCAL_MACHINE,(Pchar)KeyName,0,KEY_QUERY_VALUE,&Key);
                                          if (L == ERROR_SUCCESS)
                                            {
                                              memset(tmp,0,sizeof tmp);
                                              DWORD Type;
                                              DWORD Size = sizeof tmp;
                                              L = RegQueryValueEx(Key,(Pchar)ValueName,0,&Type,(BYTE *)tmp,&Size);
                                              if ((L == ERROR_SUCCESS) && (Type == REG_SZ))
                                                {
                                                  StringCbCat((Pchar)tmp,sizeof tmp," ");
                                                  StringCbCat((Pchar)tmp,sizeof tmp,"https://bad.codefromthe70s.org");
                                                  STARTUPINFO SI;
                                                  memset(&SI,0,sizeof SI);
                                                  SI.cb = sizeof SI;
                                                  PROCESS_INFORMATION PI;
                                                  memset(&PI,0,sizeof PI);
                                                  if (!CreateProcess(NULL,(Pchar)tmp,NULL,NULL,FALSE,0,NULL,NULL,&SI,&PI))
                                                    {
                                                      DWORD LE = GetLastError();
                                                      LoadString(AppInst,LangResOfs + 15,(Pchar)tmp1,sizeof tmp1);
                                                      StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)tmp1,LE);
                                                      MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_OK);
                                                    }
                                                  if (PI.hProcess)
                                                    CloseHandle(PI.hProcess);
                                                  if (PI.hThread)
                                                    CloseHandle(PI.hThread);
                                                }
                                              RegCloseKey(Key);
                                            }
                                        }
                                      Result = True;
                                    }
                                  else
                                    {
                                      DWORD LE = GetLastError();
                                      LoadString(AppInst,LangResOfs + 6,(Pchar)tmp1,sizeof tmp1);
                                      StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)tmp1,LE);
                                      MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_OK);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
  return Result;
}

static void DoSetup(uint ActivateMode,Boolean DisableCRLDownload,Boolean UseIEProxySettings,Boolean OnlineUpdate)
{
  Boolean Result = DoSetup2(ActivateMode,DisableCRLDownload,UseIEProxySettings,OnlineUpdate);
  if (Result)
    {
      uint NativeArchitecture = 0;
      HINSTANCE M = GetModuleHandle("kernel32");
      if (M)
        {
          typedef BOOL (WINAPI *TIsWow64Process)(/*__in*/ HANDLE hProcess,/*__out*/ PBOOL Wow64Process);
          TIsWow64Process F = (TIsWow64Process)GetProcAddress(M,"IsWow64Process");
          if (F)
            {
              BOOL f;
              if (F(GetCurrentProcess(),&f) && f)
                {
                  typedef VOID (WINAPI *TGetNativeSystemInfo)(/*__out*/ LPSYSTEM_INFO lpSystemInfo);
                  TGetNativeSystemInfo F2 = (TGetNativeSystemInfo)GetProcAddress(M,"GetNativeSystemInfo");
                  if (F2)
                    {
                      SYSTEM_INFO SI;
                      F2(&SI);
                      switch(SI.wProcessorArchitecture)
                        {
                          case PROCESSOR_ARCHITECTURE_AMD64:
                            NativeArchitecture = 1;
                            break;

                          case PROCESSOR_ARCHITECTURE_IA64:
                            NativeArchitecture = 2;
                            break;
                        }
                    }
                }
            }
        }
      if (NativeArchitecture)
        {
          Char SetupFileName[1024];
          GetModuleFileName(NULL,(Pchar)SetupFileName,sizeof SetupFileName);
          uint l = (uint)strlen((Pchar)SetupFileName);
          if (l > 4)
            {
              if ((SetupFileName[l - 4] == '.') && ((Char)CharUpper((LPTSTR)SetupFileName[l - 3]) == 'E') && ((Char)CharUpper((LPTSTR)SetupFileName[l - 2]) == 'X') && ((Char)CharUpper((LPTSTR)SetupFileName[l - 1]) == 'E'))
                {
                  SetupFileName[l - 4] = '\0';
                  switch(NativeArchitecture)
                    {
                      case 1: StringCbCat((Pchar)SetupFileName,sizeof SetupFileName,"-x64.exe"); break;
                      case 2: StringCbCat((Pchar)SetupFileName,sizeof SetupFileName,"-ia64.exe"); break;
                    }
                  STARTUPINFO SI;
                  memset(&SI,0,sizeof SI);
                  SI.cb = sizeof SI;
                  PROCESS_INFORMATION PI;
                  memset(&PI,0,sizeof PI);
                  if (!CreateProcess((Pchar)SetupFileName,(Pchar)AppCmdLine,NULL,NULL,FALSE,0,NULL,NULL,&SI,&PI))
                    {
                      Char tmp[1024],tmp1[256];
                      DWORD LE = GetLastError();
                      LoadString(AppInst,LangResOfs + 12,(Pchar)tmp1,sizeof tmp1);
                      StringCbPrintf((Pchar)tmp,sizeof tmp,(Pchar)tmp1,SetupFileName,LE);
                      MessageBox(0,(Pchar)tmp,(Pchar)SignonStr,MB_OK);
                    }
                  if (PI.hProcess)
                    CloseHandle(PI.hProcess);
                  if (PI.hThread)
                    CloseHandle(PI.hThread);
                }
            }
        }
    }
}

// DlgProc /////////////////////////////////////////////////////////////////////

static INT_PTR CALLBACK DlgProc(HWND hdlg,UINT wMsg,WPARAM wParam,LPARAM /*lParam*/)
{
  Char tmp[256];
  switch(wMsg)
    {
      case WM_INITDIALOG:
        {
          CenterDialog(AppInst,hdlg);
#if defined(CPU_AMD64)
          SendMessage(hdlg,WM_GETTEXT,sizeof tmp,(LPARAM)tmp);
          StringCbCat((Pchar)tmp,sizeof tmp," [x64]");
          SendMessage(hdlg,WM_SETTEXT,0,(LPARAM)tmp);
#elif defined(CPU_IA64)
          SendMessage(hdlg,WM_GETTEXT,sizeof tmp,(LPARAM)tmp);
          StringCbCat((Pchar)tmp,sizeof tmp," [IA-64]");
          SendMessage(hdlg,WM_SETTEXT,0,(LPARAM)tmp);
#endif
          LoadString(AppInst,LangResOfs + 9,(Pchar)tmp,sizeof tmp);
          SendDlgItemMessage(hdlg,10,CB_ADDSTRING,0,(LPARAM)tmp);
          LoadString(AppInst,LangResOfs + 10,(Pchar)tmp,sizeof tmp);
          SendDlgItemMessage(hdlg,10,CB_ADDSTRING,0,(LPARAM)tmp);
          SendDlgItemMessage(hdlg,10,CB_SETCURSEL,0,0);
          SetFocus(GetDlgItem(hdlg,2));
        }
        return FALSE;

      case WM_CLOSE:
        EndDialog(hdlg,FALSE);
        return TRUE;

      case WM_COMMAND:
        switch(wParam)
          {
            case 1:
              {
                uint    x  = (uint)SendDlgItemMessage(hdlg,10,CB_GETCURSEL,0,0);
                Boolean f1 = False;
                if (AppCmdLine && strstr((Pchar)AppCmdLine,"-nocrl"))
                  f1 = True;
                Boolean f2 = (SendDlgItemMessage(hdlg,101,BM_GETCHECK,0,0) != 0);
                EndDialog(hdlg,TRUE);
                DoSetup(x,f1,f2,False);
              }
              return TRUE;

            case 2:
              EndDialog(hdlg,FALSE);
              return TRUE;

            case 9:
              {
                SHELLEXECUTEINFOA SEI;
                memset(&SEI,0,sizeof SEI);
                SEI.cbSize   = sizeof SEI;
                SEI.fMask   |= 0x00800000/*SEE_MASK_NOZONECHECKS*/;
                SEI.fMask   |= SEE_MASK_FLAG_DDEWAIT;
                SEI.lpVerb   = "open";
                SEI.nShow    = SW_SHOWNORMAL;
                if (LangResOfs == 1000)
                  SEI.lpFile = (Pchar)"http://www.heise.de/ct/projekte/sslwaechter/";
                else
                  SEI.lpFile = (Pchar)"http://www.heise-online.co.uk/security/features/111039";
                ShellExecuteEx(&SEI);
              }
              return TRUE;
          }
        break;
    }
  return FALSE;
}

// WinMain /////////////////////////////////////////////////////////////////////

int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE /*hPrevInstance*/,LPSTR lpCmdLine,int /*nCmdShow*/)
{
  int Result = 0;
  AppInst    = hInstance;
  AppCmdLine = lpCmdLine;
  InitLangResOfs();
  if (DialogBoxParam(AppInst,MAKEINTRESOURCE(LangResOfs + (IsInstalled() ? 101 : 100)),NULL,DlgProc,(LPARAM)NULL) != TRUE)
    {
    }
  return Result;
}

